@@ -9,6 +9,7 @@
 #' @param data A data frame or list containing the model response variable and covariates required by the formula.
 #'             By default the variables are taken from environment(formula): typically the environment from which gam is called.
 #' @param qu The quantile of interest. Should be in (0, 1).
+#' @param discrete If TRUE then covariate discretisation is used for faster model fitting. See \code{mgcv::}\link[mgcv]{bam} for details.
 #' @param err An upper bound on the error of the estimated quantile curve. Should be in (0, 1). 
 #'            Since qgam v1.3 it is selected automatically, using the methods of Fasiolo et al. (2017).
 #'            The old default was \code{err=0.05}.
@@ -61,14 +62,19 @@
 #'                   \item{\code{lsig} = a vector containing the values of log(sigma) that minimize the loss function, 
 #'                                       for each quantile.}
 #'                   \item{\code{err} = the error bound used for each quantile. Generally each entry is identical to the
-#'                                      argument \code{err}, but in some cases the function increases it to enhance stabily.}
+#'                                      argument \code{err}, but in some cases the function increases it to enhance stability.}
 #'                   \item{\code{ranges} = the search ranges by the Brent algorithm to find log-sigma, for each quantile. }
 #'                   \item{\code{store} = a list, where the i-th entry is a matrix containing all the locations (1st row) at which
 #'                                        the loss function has been evaluated and its value (2nd row), for the i-th quantile.}
+#'                   \item{\code{final_fit} = a list, where the i-th entry is a list with the estimated conditional quantile 
+#'                                             (\code{mustart}), smoothing parameters (\code{in.out$sp}) and scale parameter 
+#'                                             (\code{in.out$scale}, should always be equal to 1) at the optimal value of log(sigma).}
 #' }
 #' @author Matteo Fasiolo <matteo.fasiolo@@gmail.com>. 
-#' @references Fasiolo, M., Goude, Y., Nedellec, R. and Wood, S. N. (2017). Fast calibrated additive quantile regression. Available at
-#'             \url{https://arxiv.org/abs/1707.03307}.
+#' @references Fasiolo, M., Wood, S.N., Zaffran, M., Nedellec, R. and Goude, Y., 2020. 
+#'             Fast calibrated additive quantile regression. 
+#'             Journal of the American Statistical Association (to appear).
+#'             \doi{10.1080/01621459.2020.1725521}.
 #' @examples
 #' library(qgam); library(MASS)
 #' 
@@ -131,10 +137,14 @@
 #' }
 #' } 
 #'
-tuneLearnFast <- function(form, data, qu, err = NULL,
+tuneLearnFast <- function(form, data, qu, discrete = FALSE, err = NULL,
                           multicore = !is.null(cluster), cluster = NULL, ncores = detectCores() - 1, paropts = list(),
                           control = list(), argGam = NULL)
 { 
+  discrete <- .should_we_use_discrete(form = form, discrete = discrete)
+  
+  gam_name <- ifelse(discrete, "bam", "gam")
+  
   # Removing all NAs, unused variables and factor levels from data
   data <- .cleanData(.dat = data, .form = form, .drop = argGam$drop.unused.levels)
   
@@ -153,7 +163,7 @@ tuneLearnFast <- function(form, data, qu, err = NULL,
   ctrl <- list( "loss" = "calFast", "sam" = "boot", "vtype" = "m", "epsB" = 1e-5,
                 "init" = NULL, "brac" = log( c(1/2, 2) ),  "K" = 50,
                 "redWd" = 10, "tol" = .Machine$double.eps^0.25, "aTol" = 0.05, "b" = 0,
-                "gausFit" = NULL,
+                "init_qgam" = NULL,
                 "link" = "identity",
                 "verbose" = FALSE, "progress" = TRUE )
   
@@ -166,10 +176,13 @@ tuneLearnFast <- function(form, data, qu, err = NULL,
   if( !(ctrl$loss%in%c("calFast", "cal", "pin")) ) stop("control$loss should be either \"cal\", \"pin\" or \"calFast\" ")
   if( !(ctrl$sam%in%c("boot", "kfold")) ) stop("control$sam should be either \"boot\" or \"kfold\" ")
   if( (ctrl$loss=="cal") && (ctrl$sam=="kfold")  ) stop("You can't use control$sam == \"kfold\" when ctrl$loss==\"cal\" ")
+  if(discrete && ctrl$loss != "calFast"){ stop("discrete = TRUE can be used only with control$loss != \"calFast\"") }
   
   tol <- ctrl[["tol"]]
   brac <- ctrl[["brac"]]
   
+  if( length(argGam$sp) && ctrl$loss != c("calFast") ){ stop("Cannot fix smoothing parameters unless control$loss == \"calFast\".") }
+  
   # Sanity check
   if( tol > 0.1 * abs(diff(brac)) ) stop("tol > bracket_widths/10, choose smaller tolerance or larger bracket")
   
@@ -180,26 +193,14 @@ tuneLearnFast <- function(form, data, qu, err = NULL,
     wb <- lapply(1:ctrl[["K"]], function(ii) tabulate(which(tmp != ii), n)) 
   }
   
-  # Gaussian fit, used for initialization
-  if( is.formula(form) ) {
-    gausFit <- if( is.null(ctrl[["gausFit"]]) ) { 
-      do.call("gam", c(list("formula" = form, "data" = quote(data), 
-                            "family" = gaussian(link=ctrl[["link"]])), argGam)) 
-    } else { 
-      ctrl$gausFit 
-    }
-    varHat <- gausFit$sig2
-    formL <- form
-  } else {
-    gausFit <- if( is.null(ctrl[["gausFit"]]) ) { 
-      do.call("gam", c(list("formula" = form, "data" = quote(data), 
-                            "family" = gaulss(link=list(ctrl[["link"]], "logb"), b=ctrl[["b"]])), argGam)) 
-    } else { 
-      ctrl$gausFit 
-    }
-    varHat <- 1 / gausFit$fit[ , 2]^2
-    formL <- form[[1]]
+  tmp <- ctrl$init_qgam
+  if(is.null(tmp)){
+    tmp <- .init_gauss_fit(form = form, data = data, ctrl = ctrl, argGam = argGam, qu = qu, discrete = discrete)
   }
+  gausFit <- tmp$gausFit
+  formL <- tmp$formL
+  varHat <- tmp$varHat
+  initM <- tmp$initM
   
   # Order quantiles so that those close to the median are dealt with first
   oQu <- order( abs(qu - 0.5) )
@@ -209,45 +210,51 @@ tuneLearnFast <- function(form, data, qu, err = NULL,
     # We assume lam~0 and we match the variance of a symmetric (median) Laplace density with that of the Gaussian fit.
     # This is an over-estimate for extreme quantiles, but experience suggests that it's better erring on the upper side.
     tmp <- 0.5 #qu[ oQu[1] ]
-    if( !is.list(form) ){
-      isig <- log(sqrt( gausFit$sig2 * (tmp^2*(1-tmp)^2) / (2*tmp^2-2*tmp+1) ))
-    } else {
-      isig <- log(sqrt( (ctrl[["b"]]+exp(coef(gausFit)["(Intercept).1"]))^2 * (tmp^2*(1-tmp)^2) / (2*tmp^2-2*tmp+1) ))
-    }
+    isig <- log(sqrt( mean(varHat) * (tmp^2*(1-tmp)^2) / (2*tmp^2-2*tmp+1) ))
   } else {
     isig <- ctrl[["init"]]
   }
   
   # Create gam object for full data fits
-  mObj <- do.call("gam", c(list("formula" = formL, "family" = quote(elf(qu = NA, co = NA, theta = NA, link = ctrl$link)), 
-                                "data" = quote(data), "fit" = FALSE), argGam))
+  mObj <- do.call(gam_name, c(list("formula" = formL, "family" = quote(elf(qu = NA, co = NA, theta = NA, link = ctrl$link)),
+                                   "data" = quote(data), "fit" = FALSE, "discrete" = discrete), argGam))
+  
+  # Remove "sp" as it is already been fixed
+  argGam <- argGam[ names(argGam) != "sp" ]
   
   # Create gam object for bootstrap fits
-  bObj <- do.call("gam", c(list("formula" = formL, "family" = quote(elf(qu = NA, co = NA, theta = NA, link = ctrl$link)), "data" = quote(data), 
-                                "sp" = if(length(gausFit$sp)){gausFit$sp}else{NULL}, fit = FALSE), argGam))
+  bObj <- do.call(gam_name, c(list("formula" = formL, "family" = quote(elf(qu = NA, co = NA, theta = NA, link = ctrl$link)), "data" = quote(data),
+                                   "sp" = if(length(gausFit$sp)){gausFit$sp}else{NULL}, fit = FALSE, "discrete" = discrete), argGam))
   
   # Preparing bootstrap object for gam.fit3
   bObj <- .prepBootObj(obj = bObj, eps = ctrl$epsB, control = argGam$control)
   
   # Preparing reparametrization list and hide it within mObj. This will be needed by the sandwich calibration
   if( ctrl$loss == "calFast" ){
-    mObj$hidRepara <- if(is.formula(formL)) { 
-      .prepBootObj(obj = mObj, eps = NULL, control = argGam$control)[ c("UrS", "Mp", "U1") ] 
-    } else { 
-      bObj$Sl 
-    } 
+    mObj$hidRepara <- .prepBootObj(obj = mObj, eps = NULL, control = argGam$control)[ c("UrS", "Mp", "U1") ] 
   }
   
   # Create prediction design matrices for each bootstrap sample or CV fold
   class( mObj ) <- c("gam", "glm", "lm") 
   mObj$coefficients <- rep(0, ncol(mObj$X))  # Needed to fool predict.gam
-  pMat <- predict.gam(mObj, newdata = data, type = "lpmatrix")
   
   # Stuff needed for the sandwich estimator
-  sandStuff <- list("XFull" = pMat,
-                    "EXXT" = crossprod(pMat, pMat) / n,                    # E(xx^T)
-                    "EXEXT" = tcrossprod( colMeans(pMat), colMeans(pMat))) # E(x)E(x)^T
-        
+  if(discrete){
+    pMat <- NULL
+    colmX <- XWyd(X=mObj$Xd,w=rep(1,n),y=rep(1,n),k=mObj$kd,ks=mObj$ks,
+                  ts=mObj$ts,dt=mObj$dt,v=mObj$v,qc=mObj$qc,drop=mObj$drop)/n
+    sandStuff <- list("XFull" = pMat,
+                      "EXXT" = XWXd(X=mObj$Xd,w=rep(1,n),k=mObj$kd,ks=mObj$ks,
+                                    ts=mObj$ts,dt=mObj$dt,v=mObj$v,qc=mObj$qc,
+                                    drop=mObj$drop)/n,  # E(xx^T)
+                      "EXEXT" = tcrossprod( colmX, colmX)) # E(x)E(x)^T
+  }else{
+    pMat <- predict.gam(mObj, newdata = data, type = "lpmatrix")
+    sandStuff <- list("XFull" = pMat,
+                      "EXXT" = crossprod(pMat, pMat) / n,                    # E(xx^T)
+                      "EXEXT" = tcrossprod( colMeans(pMat), colMeans(pMat))) # E(x)E(x)^T
+  }
+  
   if( multicore ){ 
     # Create cluster
     tmp <- .clusterSetUp(cluster = cluster, ncores = ncores) #, exportALL = TRUE)
@@ -263,7 +270,7 @@ tuneLearnFast <- function(form, data, qu, err = NULL,
     paropts[[".packages"]] <- NULL
     
     # Export bootstrap objects, prediction matrix and user-defined stuff
-    tmp <- unique( c("bObj", "pMat", "wb", "ctrl", "argGam", ".getVp", ".egamFit", ".gamlssFit", paropts[[".export"]]) )
+    tmp <- unique( c("bObj", "pMat", "wb", "ctrl", "argGam", ".egamFit", paropts[[".export"]]) )
     clusterExport(cluster, tmp, envir = environment())
     paropts[[".export"]] <- NULL
   }
@@ -272,12 +279,13 @@ tuneLearnFast <- function(form, data, qu, err = NULL,
   sigs <- efacts <- errors <- numeric(nq)
   rans <- matrix(NA, nq, 2)
   store <- vector("list", nq)
+  final_fit <- vector("list", nq)
   names(sigs) <- names(errors) <- rownames(rans) <- qu
   
   # Here we need bTol > aTol, otherwise the new bracket will be too close to the probable solution
   bTol <- 4*ctrl$aTol
   
-  if( is.null(err) ){ err <- .getErrParam(qu = qu, gFit = gausFit) }
+  if( is.null(err) ){ err <- .getErrParam(qu = qu, gFit = gausFit, varHat = varHat) }
   
   if(ctrl$progress){ cat("Estimating learning rate. Each dot corresponds to a loss evaluation. \n") }
   for(ii in 1:nq)
@@ -295,12 +303,13 @@ tuneLearnFast <- function(form, data, qu, err = NULL,
       
       # Estimate log(sigma) using brent methods with current bracket (srange)
       res  <- .tuneLearnFast(mObj = mObj, bObj = bObj, pMat = pMat, sandStuff = sandStuff, wb = wb, qu = qu[oi], err = err[oi],
-                             srange = srange, gausFit = gausFit, varHat = varHat,
+                             srange = srange, gausFit = gausFit, varHat = varHat, initM = initM[[oi]],
                              multicore = multicore, cluster = cluster, ncores = ncores, paropts = paropts,  
                              control = ctrl, argGam = argGam)  
       
       # Store loss function evaluations
       store[[oi]] <- cbind(store[[oi]], res[["store"]])
+      final_fit[[oi]] <- res[["final_fit"]]
       lsig <- res$minimum
       
       # If solution not too close to boundary store results and determine bracket for next iteration
@@ -360,7 +369,7 @@ tuneLearnFast <- function(form, data, qu, err = NULL,
   
   names(sigs) <- qu
   
-  out <- list("lsig" = sigs, "err" = errors, "ranges" = rans, "store" = store)
+  out <- list("lsig" = sigs, "err" = errors, "ranges" = rans, "store" = store, "final_fit" = final_fit)
   attr(out, "class") <- "learnFast"
   
   # Close the cluster if it was opened inside this function
@@ -373,24 +382,11 @@ tuneLearnFast <- function(form, data, qu, err = NULL,
 ### Internal version, which works for a single quantile qu
 ########################################################################## 
 .tuneLearnFast <- function(mObj, bObj, pMat, sandStuff, wb, qu, err,
-                           srange, gausFit, varHat,
+                           srange, gausFit, varHat, initM,
                            multicore, cluster, ncores, paropts, 
                            control, argGam)
 {
 
-  # Initializing smoothing parameters using gausFit is a very BAD idea
-  if( is.formula(mObj$formula) ) { # Extended Gam OR ...
-    coefGau <- coef(gausFit)
-    if( is.list(gausFit$formula) ){ 
-      lpi <- attr(predict(gausFit, newdata = gausFit$model[1:2, , drop = FALSE], type = "lpmatrix"), "lpi")
-      coefGau <- coefGau[ lpi[[1]] ] 
-      }
-    initM <- list("start" = coefGau + c(quantile(residuals(gausFit, type="response"), qu), rep(0, length(coefGau)-1)), 		
-                  "in.out" = NULL) # let gam() initialize sp via initial.spg() 		
-  } else { # ... GAMLSS		
-    initM <- list("start" = NULL, "in.out" = NULL) # I have no clue
-  }
-
   init <- list("initM" = initM, "initB" = vector("list", control$K))
   
   # If we get convergence error, we increase "err" up to 0.2. If the error persists (or if the 
